package cz.upol.scripting;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptException;

import cz.upol.scripting.game.CollisionResolution;
import cz.upol.scripting.game.Element;
import cz.upol.scripting.game.World;
	
/**
 * Fasada zpristupnujici funkce pro skriptovaci jazyk
 */
public class GameAPI {
	
	private static int lastId = 0;

	/** priznak reseni kolize: element se nebude pohybovat v horizontalnim smeru, ale pouze vertikalne */
	public final int DONT_MOVE_HORIZONTALLY = CollisionResolution.DONT_MOVE_HORIZONTALLY;
	
	/** priznak reseni kolize: element se nebude pohybovat ve vertikalnim smeru, ale pouze horizontalne */
	public final int DONT_MOVE_VERTICALLY = CollisionResolution.DONT_MOVE_VERTICALLY;
	
	/** priznak reseni kolize: element se nebude pohybovat vubec */
	public final int DONT_MOVE = CollisionResolution.DONT_MOVE;
	
	/** 
	 * priznak reseni kolize: algoritmus se detekovat kolize znovu,
	 * jelikoz doslo ke zmene nejakeho elementu v hernim svete 
	 */
	public final int RETRY = CollisionResolution.RETRY;
	
	private final WorldView worldView;
	private final World world;
	private final ScriptEngine scriptEngine;

	/**
	 * Vytvori fasadu pro skriptovaci jazyk
	 * 
	 * @param world -- objekt predstavujici stav hry
	 * @param worldView -- objekt predstavujici graficke zobrazeni hry
	 * @param scriptEngine -- pouzity skriptovaci engine
	 */
	public GameAPI(World world, WorldView worldView, ScriptEngine scriptEngine) {
		this.world = world;
		this.worldView = worldView;
		this.scriptEngine = scriptEngine;
	}
	
	/**
	 * Vytvori a vrati novy herni element a jeho graficke zobrazeni.
	 * Kazdy objekt je definovan svou tridou (ktera je pouzita i v CSS),
	 * souradnicemi x, y, vyskou a sirkou.
	 * 
	 * Kazdemu objektu je prirazeno pseudo-unikatni cislo.
	 */
	public Element createElement(String elementClass, double x, double y, double width, double height) {
		String id = "$" + elementClass + "$" + (++lastId);
		Element elt = new Element(id, elementClass, x, y, width, height);
		ElementView eltView = new ElementView(elt, elementClass);
		world.add(elt);
		worldView.add(eltView);
		return elt;
	}
	
	/**
	 * Vytvori a vrati novy herni element a jeho graficke zobrazeni.
	 * Kazdy objekt je definovan svym id, svou tridou (ktera je pouzita i v CSS),
	 * souradnicemi x, y, vyskou a sirkou.
	 */
	public Element createElement(String elementClass, String id, double x, double y, double width, double height) {
		Element elt = new Element(id, elementClass, x, y, width, height);
		ElementView eltView = new ElementView(elt, elementClass);
		world.add(elt);
		worldView.add(eltView);
		return elt;
	}
	
	/**
	 * Vrati element herniho sveta podle jeho id
	 */
	public Element getById(String id) {
		return world.get(id);
	}
	
	/**
	 * Vrati elementy herniho sveta, ktere maji stejnou tridu
	 */
	public Collection<Element> getByClass(String elementClass) {
		List<Element> result = new ArrayList<>();
		for (Element e : world) {
			if (e.getElementClass().equals(elementClass))
				result.add(e);
		}
		return result;
	}
	
	/**
	 * Zahaji novou hru
	 */
	public void startNewGame() {
		world.clear();
		worldView.clear();
		world.setGameComplete(null);
		try {
			((Invocable) scriptEngine).invokeFunction("onInit");
		} catch (NoSuchMethodException | ScriptException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Ukonci hru s danou zpravou (napr. win/game over)
	 */
	public void completeGame(String message) {
		world.setGameComplete(message);
	}
}