package cz.upol.scripting;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import cz.upol.scripting.game.Element;
import cz.upol.scripting.game.World;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.event.Event;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.scene.text.Font;
import javafx.stage.Stage;
import javafx.util.Duration;

public class MainApp extends Application {
	
	/** casovy usek (v ms), po kterem jsou provadeny aktualizace */
	public static final int TIME_SLICE = 100;
	
	private final String STYLE_FILE = "/style.css";
	private final String GAME_FILE = "/level01.js";
//	private final String GAME_FILE = "/level02.js";
	private final String SCRIPT_ENGINE = "JavaScript";
//	private final String GAME_FILE = "/level03.py";
//	private final String SCRIPT_ENGINE = "python";
	
	
	private final World world;
	private final WorldView worldView;
	private final GameAPI game;
	
	private final BorderPane codePane = new BorderPane();
	private CodeEditor codeEditor;
	private ScriptEngine scriptEngine;
	
	private VBox mainScreen;
	private Label finalMessage;
	
	/** kod, ktery je pouzit ve hre */
	private String code;
	
	/** hlavni casova smycka ridici animace */
	private Timeline mainTimeline;
	
	public MainApp() {
		// vytvori instanci herniho sveta, pricemz reseni kolizi je presunuto
		// do volani procedury onCollision ve skriptovacim jazyce
		world = new World() {

			@Override
			protected int resolveCollisions(Element elt, Element obstackle, boolean horizontalCollision,
					boolean verticalCollision) {
				try {
					Invocable invocable = (Invocable) scriptEngine;
					return (int) invocable.invokeFunction("onCollision", elt, obstackle, horizontalCollision, verticalCollision);
				} catch (NoSuchMethodException | ScriptException e) {
					e.printStackTrace();
					return 0;
				}
			}
		};
		
		worldView = new WorldView();
	
		// vytovri skriptovaci engine
		ScriptEngineManager sem = new ScriptEngineManager();
		scriptEngine = sem.getEngineByName(SCRIPT_ENGINE);
		
		// vytvori fasadu/API pro skriptovaci jazyk
		game = new GameAPI(world, worldView, scriptEngine);
		
		// predani odkazu do skriptovaciho enginu, resp. jazyka
		scriptEngine.put("game", game);
	}
	
	/**
	 * Reakce na stisknute klavesy.
	 * F4 -- otevreni editoru
	 * Ctl+N -- nova hra
	 * 
	 * Ostatni klavesy presmerovany do procedury <tt>onKeyPressed</tt>
	 * skriptovaciho jazyka.
	 */
	private void dispatchKeyEvents(KeyEvent e) {
		if (e.getCode() == KeyCode.F4) {
			String newCode = codeEditor.open(code);
			if (newCode != null) {
				installCode(newCode); 
				game.startNewGame();
			}
		} else if ((e.getCode() == KeyCode.N) && (e.isControlDown())){
			game.startNewGame();
		} else {
			Invocable invocable = (Invocable) scriptEngine;
			try {
				invocable.invokeFunction("onKeyPressed", e.getCode());
			} catch (NoSuchMethodException | ScriptException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	/**
	 * Hlavni procedura z pohledu JavaFX, spusteni hry
	 */
	@Override
	public void start(Stage primaryStage) throws Exception {
		
		codeEditor = new CodeEditor(primaryStage);
		
		installCode(loadSource(GAME_FILE));
		game.startNewGame();
		
		createMainLoop();
		
		Pane root = createLayout();
		
		Scene scene = new Scene(root, 500, 350);
		scene.getStylesheets().addAll(this.getClass().getResource(STYLE_FILE).toExternalForm());
		scene.setOnKeyPressed(this::dispatchKeyEvents);
		
		primaryStage.setTitle("Game");
		primaryStage.setScene(scene);
		primaryStage.show();
	}
	
	/**
	 * Vytvori rozlozeni grafickych prvku aplikace
	 */
	private Pane createLayout() {
		mainScreen = new VBox(30);
		mainScreen.setAlignment(Pos.CENTER);
		
		Button startBtn = new Button("Start");
		startBtn.setOnAction(e -> {
			hideMainScreen();
			game.startNewGame();
		});
		
		finalMessage = new Label("Welcome");
		finalMessage.setFont(Font.font(40));
		worldView.setOpacity(0.4);
		
		mainScreen.getChildren().addAll(finalMessage, startBtn);
		StackPane mainPane = new StackPane(worldView, mainScreen);
		
		BorderPane root = new BorderPane();
		root.setCenter(mainPane);
		root.setRight(codePane);
		
		return root;
	}
	
	/**
	 * Zobrazi hlavni obrazovku s jednoduchou zpravou
	 * a tlacitkem start.
	 */
	private void showMainScreen(String message) {
		mainTimeline.stop();
		worldView.setOpacity(0.4);
		mainScreen.setVisible(true);
		finalMessage.setText(message);
	}
	
	/**
	 * Skryje hlavni obrazovku a spusti hru.
	 */
	private void hideMainScreen() {
		worldView.setOpacity(1);
		mainScreen.setVisible(false);
		mainTimeline.play();
	}

	/**
	 * Vytvori herni smycku, ktera pravidelne aktualizuje
	 * herni svet i jeho grafickou reprezentaci
	 */
	private void createMainLoop() {
		mainTimeline = new Timeline();
		KeyFrame updates = new KeyFrame(Duration.millis(TIME_SLICE), this::update);
		
		mainTimeline.getKeyFrames().add(updates);
		mainTimeline.setCycleCount(Animation.INDEFINITE);
	}
	
	/**
	 * Provede aktulizaci herniho sveta a to v nasledujicich krocich.
	 * (1) Posune vsechny objekty v hernim sveta a vyresi jejich kolize.
	 * (2) Provede animaci posunu, prip. dalsich zmen.
	 * (3) Odstrani z herniho sveta zmizele objekty.
	 * (4) Zavola proceduru <tt>onLoop()</tt> ve skriptovacim enginue/jazyce,
	 * a provede pripadne dalsi zmeny.
	 */
	private void update(Event e) {
		world.move(); 
        worldView.singleUpdate();
		world.update();
		Invocable invocable = (Invocable) scriptEngine;
		try {
			invocable.invokeFunction("onLoop");
		} catch (NoSuchMethodException | ScriptException ex) {
			ex.printStackTrace();
		}
		if (world.getGameComplete() != null) {
			showMainScreen(world.getGameComplete());
		}
			
	}

	/**
	 * Nahraje z daneho zdroje zdrojovy kod skriptu 
	 */
	private String loadSource(String fileName) throws IOException {
		StringBuilder codeBuf = new StringBuilder();
		try (BufferedReader rd = new BufferedReader(new InputStreamReader(this.getClass().getResource(fileName).openStream()))) {
			while (rd.ready())
				codeBuf.append(rd.readLine()).append("\n");
		}
		return codeBuf.toString();
	}
	
	/**
	 * Provede inicializaci dle zadaneho kodu
	 */
	private void installCode(String code) {
		this.code = code;
		try {
			scriptEngine.eval(code);
		} catch (ScriptException e) {
			e.printStackTrace();
		}
	}
	
	public static void main(String[] args) {
		launch(args);
	}

}
