package cz.upol.scripting.game;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Predstavuje horni svet jako kolekci elementu ({@link Element})
 * herniho sveta.
 */
public class World implements Iterable<Element>, CollisionResolution {
	
	private Map<String, Element> elements = new HashMap<>();
	
	/** 
	 * Priznak, zda je hra aktivni, nebo ne.
	 * Viz {@link World#getGameComplete()}.
	 */
	private String gameComplete = null;
	
	/**
	 * Vlozi novy element do horniho sveta
	 */
	public void add(Element element) {
		elements.put(element.getId(), element);
	}
	
	/**
	 * Odstrani z herniho sveta dany element
	 */
	public void remove(Element element) {
		elements.remove(element.getId());
	}
	
	/**
	 * Odstrani vsechny elementy z herniho sveta
	 */
	public void clear() {
		elements.clear();
	}
	
	/**
	 * @return vraci element s konkretnim id
	 */
	public Element get(String id) {
		return elements.get(id);
	}
	
	/** 
	 * @return vraci priznak, zda je hra aktivni, nebo ne.
	 * Pokud <tt>(gameComplete == null)</tt> znamena, ze hra probiha,
	 * jinak hra neprobiha a muze zobrazit danou zpravu uzivateli.
	 */
	public String getGameComplete() {
		return gameComplete;
	}

	/**
	 * Nastavi priznak zda je hra aktivni, viz {@link World#getGameComplete()}
	 */
	public void setGameComplete(String gameComplete) {
		this.gameComplete = gameComplete;
	}
	
	/**
	 * Posune vsechny elementy ve smeru danem jejich rychlostmi.
	 * Pokud dojde ke kolizi, snazi se ji vyresit. 
	 */
	public void move() {
		Element elt = null;
		boolean retry = false;
		Iterator<Element> it = elements.values().iterator();
	
		while (retry || it.hasNext()) {
			if (!retry) elt = it.next();
			retry = false;
			
			Map<Element, CollisionType> collisions = findCollisions(elt);
			if (collisions == null)
				elt.move();
			else {
				int resolution = findResolution(elt, collisions);
				
				if (resolution == 0) elt.move();
				if (resolution == DONT_MOVE_HORIZONTALLY) elt.moveVertically();
				if (resolution == DONT_MOVE_VERTICALLY) elt.moveHorizontally();
				if (resolution == RETRY) retry = true;
			}
		}
	}
	
	/**
	 * Najde vsechny elementy, ktery by posunutim daneho elementu (v jim danem smeru)
	 * zacaly kolidavat s danym objektem. Rozlisuje se, zda budou spolu kolidovat 
	 * vertikalni, horizontalne, ci v obou smerech, viz {@link CollisionType}.
	 */
	private Map<Element, CollisionType> findCollisions(Element elt) {
		Map<Element, CollisionType> collisions = null;
		Rect vMove = elt.getRectBounds().move(new Vector(0, elt.getSpeed().getY()));
		Rect hMove = elt.getRectBounds().move(new Vector(elt.getSpeed().getX(), 0));
		for (Element e : elements.values()) {
			if (e == elt) continue;
			boolean vOverlap = vMove.isOverlapping(e.getRectBounds());
			boolean hOverlap = hMove.isOverlapping(e.getRectBounds());
			
			if ((hOverlap || vOverlap) && (collisions == null))
				collisions = new HashMap<>();
			
			if (vOverlap && hOverlap) collisions.put(e, CollisionType.HORIZONTAL_VERTICAL);
			else if (vOverlap) collisions.put(e, CollisionType.VERTICAL);
			else if (hOverlap) collisions.put(e, CollisionType.HORIZONTAL);
		}
		return collisions;
	}

	/**
	 * Pro dany element (<tt>elt</tt>) a s nim kolidujici elementy (<tt>collisions</tt>)
	 * se snazi najit reseni, jak se vyporadat s kolizi.
	 * Mozna reseni jsou dana jako bitvoy soucet hodnot z {@link CollisionResolution}.
	 */
	private int findResolution(Element elt, Map<Element, CollisionType> collisions) {
		int resolution = 0;
		for (Entry<Element, CollisionType> col : collisions.entrySet()) {
			CollisionType colType = col.getValue();
			boolean verticalCollision = (colType == CollisionType.HORIZONTAL_VERTICAL) || (colType == CollisionType.VERTICAL);
			boolean horizontalCollision = (colType == CollisionType.HORIZONTAL_VERTICAL) || (colType == CollisionType.HORIZONTAL);
			resolution |= resolveCollisions(elt, col.getKey(), horizontalCollision, verticalCollision); 
		}
		return resolution;
	}

	/**
	 * @return pro dany element a s nim kolidujici prekazku vraci reseni kolize,
	 * ktere je vybrano z hodnot  {@link CollisionResolution}.
	 * 
	 * Predpoklada se, ze tato metoda bude prekryta v potomkovi, protoze implicitne
	 * kolizi nijak neresi.
	 * 
	 * @param elt -- element
	 * @param obstacle -- kolidujici prekazka
	 * @param horizontalCollision -- priznak, zda elementy koliduji v horizontalnim smeru
	 * @param verticalCollision -- priznak, zda elementy koliduji v horizontalnim smeru
	 */
	protected int resolveCollisions(Element elt, Element obstacle, boolean horizontalCollision, boolean verticalCollision) {
		return 0;
	}
	
	/**
	 * Odstrani z herniho sveta vsechny elementy,
	 * ktere maji nastaveny priznak {@link Element#isDisappering()}.
	 */
	public void update() {
		Set<String> toRemove = new HashSet<>();
		for (Element elt : elements.values())
			if (elt.isDisappering())
				toRemove.add(elt.getId());
		
		for (String id : toRemove)
			elements.remove(id);
	}

	@Override
	public Iterator<Element> iterator() {
		return elements.values().iterator();
	}
}
